@@ -23,12 +23,14 @@ Utilities with minimum-depends for use in setup.py
 import email
 import os
 import re
-import StringIO
 import subprocess
 import sys
 
+from d2to1.extern import six
 from distutils.command import install as du_install
 from distutils import log
+import pkg_resources
+from setuptools.command import easy_install
 from setuptools.command import install
 from setuptools.command import sdist
 
@@ -59,7 +61,7 @@ def canonicalize_emails(changelog, mapping):
     """Takes in a string and an email alias mapping and replaces all
        instances of the aliases in the string with their real email.
     """
-    for alias, email_address in mapping.iteritems():
+    for alias, email_address in six.iteritems(mapping):
         changelog = changelog.replace(alias, email_address)
     return changelog
 
@@ -75,20 +77,33 @@ def get_reqs_from_files(requirements_files):
 
 def parse_requirements(requirements_files=['requirements.txt',
                                            'tools/pip-requires']):
+
+    def egg_fragment(match):
+        # take a versioned egg fragment and return a
+        # versioned package requirement e.g.
+        # nova-1.2.3 becomes nova>=1.2.3
+        return re.sub(r'([\w.]+)-([\w.-]+)',
+                      r'\1>=\2',
+                      match.group(1))
+
     requirements = []
     for line in get_reqs_from_files(requirements_files):
         # For the requirements list, we need to inject only the portion
         # after egg= so that distutils knows the package it's looking for
         # such as:
         # -e git://github.com/openstack/nova/master#egg=nova
+        # -e git://github.com/openstack/nova/master#egg=nova-1.2.3
         if re.match(r'\s*-e\s+', line):
-            requirements.append(re.sub(r'\s*-e\s+.*#egg=(.*)$', r'\1',
-                                line))
+            requirements.append(re.sub(r'\s*-e\s+.*#egg=(.*)$',
+                                       egg_fragment,
+                                       line))
         # such as:
         # http://github.com/openstack/nova/zipball/master#egg=nova
+        # http://github.com/openstack/nova/zipball/master#egg=nova-1.2.3
         elif re.match(r'\s*https?:', line):
-            requirements.append(re.sub(r'\s*https?:.*#egg=(.*)$', r'\1',
-                                line))
+            requirements.append(re.sub(r'\s*https?:.*#egg=(.*)$',
+                                       egg_fragment,
+                                       line))
         # -f lines are for index locations, and don't get used here
         elif re.match(r'\s*-f\s+', line):
             pass
@@ -136,7 +151,7 @@ def _run_shell_command(cmd, throw_on_error=False):
         return None
     if len(out[0].strip()) == 0:
         return None
-    return out[0].strip()
+    return out[0].strip().decode('utf-8')
 
 
 def _get_git_directory():
@@ -163,8 +178,9 @@ def write_git_changelog(git_dir=None, dest_dir=os.path.curdir,
             git_log_cmd = 'git --git-dir=%s log' % git_dir
             changelog = _run_shell_command(git_log_cmd)
             mailmap = read_git_mailmap(git_dir)
-            with open(new_changelog, "w") as changelog_file:
-                changelog_file.write(canonicalize_emails(changelog, mailmap))
+            with open(new_changelog, "wb") as changelog_file:
+                changelog_file.write(canonicalize_emails(
+                    changelog, mailmap).encode('utf-8'))
 
 
 def generate_authors(git_dir=None, dest_dir='.', option_dict=dict()):
@@ -194,21 +210,21 @@ def generate_authors(git_dir=None, dest_dir='.', option_dict=dict()):
                 changelog = "\n".join((changelog, new_entries))
 
             mailmap = read_git_mailmap(git_dir)
-            with open(new_authors, 'w') as new_authors_fh:
-                new_authors_fh.write(canonicalize_emails(changelog, mailmap))
+            with open(new_authors, 'wb') as new_authors_fh:
+                new_authors_fh.write(canonicalize_emails(
+                    changelog, mailmap).encode('utf-8'))
                 if os.path.exists(old_authors):
-                    with open(old_authors, "r") as old_authors_fh:
-                        new_authors_fh.write('\n' + old_authors_fh.read())
+                    with open(old_authors, "rb") as old_authors_fh:
+                        new_authors_fh.write(b'\n' + old_authors_fh.read())
 
 
 _rst_template = """%(heading)s
 %(underline)s
 
 .. automodule:: %(module)s
-
-:members:
-:undoc-members:
-:show-inheritance:
+  :members:
+  :undoc-members:
+  :show-inheritance:
 """
 
 
@@ -224,7 +240,43 @@ class DistutilsInstall(install.install):
 
     command_name = 'install'
 
+    def fetch_build_egg(self, req):
+        """Fetch an egg needed for building."""
+        try:
+            cmd = self._egg_fetcher
+            cmd.package_index.to_scan = []
+        except AttributeError:
+            dist = self.distribution.__class__(
+                {'script_args': ['easy_install']})
+            dist.parse_config_files()
+            opts = dist.get_option_dict('easy_install')
+            keep = (
+                'find_links', 'site_dirs', 'index_url', 'optimize',
+                'site_dirs', 'allow_hosts'
+            )
+            for key in opts.keys():
+                if key not in keep:
+                    del opts[key]   # don't use any other settings
+            if self.distribution.dependency_links:
+                links = self.distribution.dependency_links[:]
+                if 'find_links' in opts:
+                    links = opts['find_links'][1].split() + links
+                opts['find_links'] = ('setup', links)
+            cmd = easy_install.easy_install(
+                dist, args=["x"],
+                always_copy=False, build_directory=None, editable=False,
+                upgrade=False, multi_version=True, no_report=True
+            )
+            cmd.ensure_finalized()
+            self._egg_fetcher = cmd
+        return cmd.easy_install(req)
+
     def run(self):
+        for dist in pkg_resources.working_set.resolve(
+                pkg_resources.parse_requirements(
+                    self.distribution.install_requires),
+                installer=self.fetch_build_egg):
+            pkg_resources.working_set.add(dist)
         return du_install.install.run(self)
 
 
@@ -263,8 +315,9 @@ try:
                 os.makedirs(source_dir)
             for pkg in self.distribution.packages:
                 if '.' not in pkg:
-                    os.path.walk(pkg, _find_modules, modules)
-            module_list = modules.keys()
+                    for dirpath, dirnames, files in os.walk(pkg):
+                        _find_modules(modules, dirpath, files)
+            module_list = list(modules.keys())
             module_list.sort()
             autoindex_filename = os.path.join(source_dir, 'autoindex.rst')
             with open(autoindex_filename, 'w') as autoindex:
@@ -288,7 +341,7 @@ try:
 
         def _sphinx_run(self):
             if not self.verbose:
-                status_stream = StringIO.StringIO()
+                status_stream = six.StringIO()
             else:
                 status_stream = sys.stdout
             confoverrides = {}
